@@ -107,21 +107,17 @@
 //! ```
 use std::{
     convert::TryFrom,
-    collections::BTreeSet,
     path::Path,
     path::PathBuf,
     ops::Deref,
     ops::DerefMut,
     pin::Pin,
-    sync::Arc,
 };
-use anyhow::Result;
 
 use sequoia_openpgp as openpgp;
 use sequoia_ipc as ipc;
 use openpgp::{
     Cert,
-    Fingerprint,
     cert::ValidCert,
     crypto::{
         self,
@@ -151,13 +147,9 @@ use openpgp::{
 };
 use ipc::{
     Keygrip,
-    assuan::{self, Response, escape},
     sexp::Sexp,
 };
 
-use sequoia_cert_store as cert_store;
-use cert_store::LazyCert;
-
 use futures::stream::StreamExt;
 
 #[macro_use]
@@ -166,6 +158,12 @@ mod macros;
 mod babel;
 mod utils;
 
+pub mod assuan;
+use assuan::Response;
+use assuan::escape;
+
+pub mod gnupg;
+
 pub mod keyinfo;
 pub mod cardinfo;
 
@@ -179,6 +177,46 @@ trace_module!(TRACE);
 /// Some of our public types use types defined in this crate.
 pub use sequoia_ipc;
 
+#[derive(thiserror::Error, Debug)]
+/// Errors used in this module.
+pub enum Error {
+    /// GnuPG's home directory doesn't exist.
+    #[error("GnuPG's home directory ({0}) doesn't exist")]
+    GnuPGHomeMissing(PathBuf),
+    /// Unknown key.
+    #[error("Unknown key: {0}")]
+    UnknownKey(Keygrip),
+    /// No smartcards are connected.
+    #[error("No smartcards are connected")]
+    NoSmartcards,
+    /// The key already exists.
+    #[error("{0} already exists: {1}")]
+    KeyExists(Keygrip, String),
+
+    #[error(transparent)]
+    Io(#[from] std::io::Error),
+
+    #[error(transparent)]
+    Utf8(#[from] std::str::Utf8Error),
+
+    #[error(transparent)]
+    Assuan(#[from] assuan::Error),
+
+    #[error(transparent)]
+    GnuPG(#[from] gnupg::Error),
+
+    #[error(transparent)]
+    KeyInfo(#[from] keyinfo::Error),
+
+    #[error(transparent)]
+    OpenPGP(#[from] openpgp::Error),
+
+    #[error(transparent)]
+    Other(#[from] anyhow::Error),
+}
+
+type Result<R, E=Error> = std::result::Result<R, E>;
+
 /// Controls how gpg-agent inquires passwords.
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub enum PinentryMode {
@@ -295,12 +333,12 @@ impl Agent {
     /// If a `gpg-agent` isn't listening on the specified context,
     /// this function tries to start one.
     ///
-    /// The GnuPG context, `sequoia_ipc::gnupg::Context`, specifies
-    /// the GnuPG home directory.
-    pub async fn connect(ctx: &ipc::gnupg::Context) -> Result<Self> {
+    /// The GnuPG context, [`gnupg::Context`], specifies the GnuPG
+    /// home directory.
+    pub async fn connect(ctx: &gnupg::Context) -> Result<Self> {
         tracer!(TRACE, "connect");
 
-        async fn transaction(ctx: &ipc::gnupg::Context) -> Result<Agent> {
+        async fn transaction(ctx: &gnupg::Context) -> Result<Agent> {
             t!("Starting daemon if not running");
 
             // If the home directory doesn't exist, gpg-agent won't
@@ -362,12 +400,12 @@ impl Agent {
     /// If a `gpg-agent` isn't listening on the specified context,
     /// this function tries to start one.
     ///
-    /// A default `sequoia_ipc::gnupg::Context` with the specified
-    /// home directory is used as the context.
+    /// A default [`gnupg::Context`] with the specified home directory
+    /// is used as the context.
     pub async fn connect_to<P>(homedir: P) -> Result<Self>
         where P: AsRef<Path>
     {
-        let ctx = ipc::gnupg::Context::with_homedir(homedir)?;
+        let ctx = gnupg::Context::with_homedir(homedir)?;
         Self::connect(&ctx).await
     }
 
@@ -376,21 +414,21 @@ impl Agent {
     /// If a `gpg-agent` isn't listening on the specified context,
     /// this function tries to start one.
     ///
-    /// A default `sequoia_ipc::gnupg::Context` with the default
-    /// home directory is used as the context.
+    /// A default [`gnupg::Context`] with the default home directory
+    /// is used as the context.
     pub async fn connect_to_default() -> Result<Self>
     {
-        let ctx = ipc::gnupg::Context::new()?;
+        let ctx = gnupg::Context::new()?;
         Self::connect(&ctx).await
     }
 
     /// Connects to a new, ephemeral `gpg-agent`.
     ///
-    /// A default `sequoia_ipc::gnupg::Context` with an ephemeral
-    /// home directory is used as the context.
+    /// A default [`gnupg::Context`] with an ephemeral home directory
+    /// is used as the context.
     pub async fn connect_to_ephemeral() -> Result<Self>
     {
-        let ctx = ipc::gnupg::Context::ephemeral()?;
+        let ctx = gnupg::Context::ephemeral()?;
         Self::connect(&ctx).await
     }
 
@@ -445,23 +483,6 @@ impl Agent {
         Ok(self.send_simple(format!("HAVEKEY {}", grip)).await.is_ok())
     }
 
-    /// Returns the (sub)keys of the given cert that have a secret in
-    /// the agent.
-    pub async fn has_keys(&mut self,
-                          cert: &Arc<LazyCert<'_>>)
-                          -> Result<BTreeSet<Fingerprint>>
-    {
-        let mut result = BTreeSet::default();
-
-        for k in cert.keys() {
-            if self.has_key(&k).await.unwrap_or(false) {
-                result.insert(k.fingerprint());
-            }
-        }
-
-        Ok(result)
-    }
-
     /// Returns the keys managed by the agent.
     pub async fn list_keys(&mut self) -> Result<keyinfo::KeyInfoList> {
         self.send("KEYINFO --list")?;
@@ -535,7 +556,7 @@ impl Agent {
     ///
     /// This sends `LEARN --sendinfo` to gpg-agent.
     pub async fn card_info(&mut self) -> Result<cardinfo::CardInfo> {
-        self.send("learn --sending")?;
+        self.send("learn --sendinfo")?;
 
         let mut raw: Vec<(String, String)> = Vec::new();
         while let Some(response) = self.next().await {
@@ -751,7 +772,7 @@ impl Agent {
                         p.push(Sexp::String("0".into()));
                     },
                     S2K::Private { .. } | S2K::Unknown { .. } | _ => {
-                        return Err(anyhow::anyhow!("Unsupported protection mode"));
+                        return Err(anyhow::anyhow!("Unsupported protection mode").into());
                     },
                 }
 
@@ -760,7 +781,7 @@ impl Agent {
                     // We must omit the IV here.
                     skey.push(Sexp::String(c[iv_len.min(c.len())..].into()));
                 } else {
-                    return Err(anyhow::anyhow!("Failed to parse ciphertext"));
+                    return Err(anyhow::anyhow!("Failed to parse ciphertext").into());
                 }
 
                 Sexp::List(p)
@@ -969,7 +990,7 @@ impl Agent {
     /// adding `allow-preset-passphrase` to their `gpg-agent.conf`.
     /// If this is not set, which will usually be the case since it is
     /// not the default, this operation will fail and return
-    /// `Error::OperationFailed`.
+    /// [`assuan::Error::OperationFailed`].
     pub async fn preset_passphrase(&mut self,
                                    keygrip: &Keygrip,
                                    password: Password)
@@ -984,12 +1005,7 @@ impl Agent {
             format!("PRESET_PASSPHRASE {} -1 {}",
                     keygrip, escaped)).await
             .map_err(|err| {
-                // XXX: Remap ipc::gnupg::Error to Error.
-                if let Some(ipc::gnupg::Error::OperationFailed(x)) = err.downcast_ref() {
-                    Error::OperationFailed(x.clone()).into()
-                } else {
-                    err
-                }
+                err
             })?;
 
         Ok(())
@@ -1141,20 +1157,20 @@ impl Agent {
                                  -> Result<T>
     {
         self.operation_failed_as(
-            Error::OperationFailed(
+            assuan::Error::OperationFailed(
                 message.as_ref().map(|e| e.to_string())
                     .unwrap_or_else(|| "Unknown reason".into()))
                 .into()).await
     }
 
-    async fn operation_failed_as<T>(&mut self, err: anyhow::Error)
+    async fn operation_failed_as<T>(&mut self, err: Error)
                                     -> Result<T>
     {
         tracer!(TRACE, "operation_failed_as");
 
         if let Some(response) = self.next().await {
             t!("Got unexpected response {:?}", response);
-            Err(Error::ProtocolError(
+            Err(assuan::Error::ProtocolError(
                 format!("Got unexpected response {:?}", response))
                 .into())
         } else {
@@ -1228,6 +1244,39 @@ impl Agent {
         Ok(pair)
     }
 
+    /// Exports a secret key from the agent.
+    ///
+    /// `key` is the secret key that will be exported.
+    pub async fn export(&mut self, key: Key<PublicParts, UnspecifiedRole>)
+        -> Result<Key<SecretParts, UnspecifiedRole>>
+    {
+        let keygrip = Keygrip::of(key.mpis())?;
+
+        let kek = self.send_simple("KEYWRAP_KEY --export").await?;
+
+        let encrypted_key = self.send_simple(
+            format!("EXPORT_KEY {}", keygrip.to_string())).await?;
+
+        let secret_key = openpgp::crypto::ecdh::aes_key_unwrap(
+            openpgp::types::SymmetricAlgorithm::AES128,
+            &kek,
+            &encrypted_key.as_ref())?;
+
+        // Strip any trailing NULs.  They are only there for padding
+        // purposes.
+        let mut secret_key = secret_key.as_ref();
+        while ! secret_key.is_empty() && secret_key[secret_key.len() - 1] == 0 {
+            secret_key = &secret_key[..secret_key.len() - 1];
+        }
+
+        let sexp = Sexp::from_bytes(secret_key)?;
+
+        let secret_key = sexp.to_secret_key(Some(key.mpis()))?;
+
+        let (key, _) = key.add_secret(secret_key.into());
+
+        Ok(key)
+    }
 }
 
 /// Returns a convenient Err value for use in the state machines
@@ -1236,7 +1285,7 @@ fn protocol_error<T>(response: &Response) -> Result<T> {
     tracer!(TRACE, "operation_failed");
 
     t!("Got unexpected response {:?}", response);
-    Err(Error::ProtocolError(
+    Err(assuan::Error::ProtocolError(
         format!("Got unexpected response {:?}", response))
         .into())
 }
@@ -1311,6 +1360,7 @@ pub struct KeyPair {
     password: Option<crypto::Password>,
     pinentry_mode: Option<PinentryMode>,
     password_prompt: String,
+    delete_prompt: String,
 }
 
 impl KeyPair {
@@ -1324,12 +1374,21 @@ impl KeyPair {
             key.keyid(), Timestamp::try_from(key.creation_time()).unwrap())
     }
 
+    fn default_delete_prompt(key: &Key<PublicParts, UnspecifiedRole>)
+        -> String
+    {
+        format!(
+            "Do you really want to permanently delete the OpenPGP secret key:\n\
+             ID {:X}, created {}.",
+            key.keyid(), Timestamp::try_from(key.creation_time()).unwrap())
+    }
+
     /// Returns a `KeyPair` for `key` with the secret bits managed by
     /// the agent.
     ///
     /// This provides a convenient, synchronous interface for use with
     /// the low-level Sequoia crate.
-    pub fn new_for_gnupg_context<R>(ctx: &ipc::gnupg::Context,
+    pub fn new_for_gnupg_context<R>(ctx: &gnupg::Context,
                                     key: &Key<PublicParts, R>)
         -> Result<KeyPair>
         where R: KeyRole
@@ -1339,6 +1398,7 @@ impl KeyPair {
             password: None,
             pinentry_mode: None,
             password_prompt: Self::default_password_prompt(key),
+            delete_prompt: Self::default_delete_prompt(key),
             public: key.clone(),
             agent_socket: ctx.socket("agent")?.into(),
         })
@@ -1359,6 +1419,7 @@ impl KeyPair {
             password: None,
             pinentry_mode: None,
             password_prompt: Self::default_password_prompt(key),
+            delete_prompt: Self::default_delete_prompt(key),
             public: key.clone(),
             agent_socket: agent_socket.as_ref().to_path_buf(),
         })
@@ -1376,10 +1437,10 @@ impl KeyPair {
     pub fn with_cert(self, cert: &ValidCert) -> Self {
         let primary_id = cert.keyid();
         let keyid = self.public.keyid();
-        let prompt = match (primary_id == keyid,
-                            cert.primary_userid()
-                            .map(|uid| uid.clone())
-                            .ok())
+        let password_prompt = match (primary_id == keyid,
+                                     cert.primary_userid()
+                                     .map(|uid| uid.clone())
+                                     .ok())
         {
             (true, Some(uid)) => format!(
                 "Please enter the passphrase to \
@@ -1420,7 +1481,31 @@ impl KeyPair {
                 primary_id,
             ),
         };
-        self.with_password_prompt(prompt)
+
+        let delete_prompt = match cert.primary_userid()
+                                      .map(|uid| uid.clone())
+                                      .ok()
+        {
+            Some(uid) => format!(
+                "Do you really want to permanently delete the OpenPGP secret key:\n\
+                 {}\n\
+                 ID {:X}, created {}.",
+                uid.userid(),
+                keyid,
+                Timestamp::try_from(self.public.creation_time())
+                    .expect("creation time is representable"),
+            ),
+            None => format!(
+                "Do you really want to permanently delete the OpenPGP secret key:\n\
+                 ID {:X}, created {}.",
+                keyid,
+                Timestamp::try_from(self.public.creation_time())
+                    .expect("creation time is representable"),
+            ),
+        };
+
+        self.with_password_prompt(password_prompt)
+            .with_delete_prompt(delete_prompt)
     }
 
     /// Supplies a password to unlock the secret key.
@@ -1464,6 +1549,92 @@ impl KeyPair {
         self.password_prompt = prompt;
         self
     }
+
+    /// Changes the delete prompt.
+    ///
+    /// Use this function to give more context to the user when she is
+    /// prompted to delete a key.
+    ///
+    /// To set an password prompt that uses information from the
+    /// OpenPGP certificate, use [`KeyPair::with_cert`].
+    pub fn with_delete_prompt(mut self, prompt: String) -> Self {
+        self.delete_prompt = prompt;
+        self
+    }
+
+    /// Changes the key's password.
+    ///
+    /// If `preset_password` is true, then the password will also be
+    /// added to the password cache.
+    pub async fn password(&mut self, preset_password: bool)
+        -> Result<()>
+    {
+        let keygrip = Keygrip::of(self.public.mpis())?;
+
+        // Connect to the agent.
+        let mut agent = Agent::connect_to_agent(&self.agent_socket).await?;
+
+        for option in Agent::options() {
+            agent.send_simple(option).await?;
+        }
+
+        if let Some(mode) = agent.pinentry_mode.clone() {
+            agent.send_simple(
+                format!("OPTION pinentry-mode={}", mode.as_str())).await?;
+        }
+        agent.send_simple(
+            format!("SETKEYDESC {}",
+                    assuan::escape(&self.password_prompt))).await?;
+
+        agent.send_simple(
+            format!("PASSWD {}{}",
+                    if preset_password {
+                        "--preset "
+                    } else {
+                        ""
+                    },
+                    keygrip.to_string()))
+            .await?;
+
+        Ok(())
+    }
+
+    /// Deletes the specified key.
+    ///
+    /// If `stub_only` is true, then the key will be removed if it is
+    /// a stub that corresponds to a key on a token.
+    pub async fn delete_key(&mut self, stub_only: bool)
+        -> Result<()>
+    {
+        let keygrip = Keygrip::of(self.public.mpis())?;
+
+        // Connect to the agent.
+        let mut agent = Agent::connect_to_agent(&self.agent_socket).await?;
+
+        for option in Agent::options() {
+            agent.send_simple(option).await?;
+        }
+
+        if let Some(mode) = agent.pinentry_mode.clone() {
+            agent.send_simple(
+                format!("OPTION pinentry-mode={}", mode.as_str())).await?;
+        }
+        agent.send_simple(
+            format!("SETKEYDESC {}",
+                    assuan::escape(&self.delete_prompt))).await?;
+
+        agent.send_simple(
+            format!("DELETE_KEY {}{}",
+                    if stub_only {
+                        "--stub-only "
+                    } else {
+                        ""
+                    },
+                    keygrip.to_string()))
+            .await?;
+
+        Ok(())
+    }
 }
 
 impl KeyPair {
@@ -1517,7 +1688,7 @@ impl KeyPair {
                         _ => agent.acknowledge_inquiry().await?,
                     },
                 assuan::Response::Error { ref message, .. } =>
-                    return agent.operation_failed(message).await,
+                    return Ok(agent.operation_failed(message).await?),
                 assuan::Response::Data { ref partial } =>
                     data.extend_from_slice(partial),
             }
@@ -1614,7 +1785,7 @@ impl KeyPair {
                         padding = message != "0";
                     },
                 assuan::Response::Error { ref message, .. } =>
-                    return agent.operation_failed(message).await,
+                    return Ok(agent.operation_failed(message).await?),
                 assuan::Response::Data { ref partial } =>
                     data.extend_from_slice(partial),
             }
@@ -1675,32 +1846,6 @@ fn map_panic(_: Box<dyn std::any::Any + std::marker::Send>) -> anyhow::Error
     anyhow::anyhow!("worker thread panicked")
 }
 
-#[derive(thiserror::Error, Debug)]
-/// Errors used in this module.
-pub enum Error {
-    /// Errors related to `gpgconf`.
-    #[error("gpgconf: {0}")]
-    GPGConf(String),
-    /// The remote operation failed.
-    #[error("Operation failed: {0}")]
-    OperationFailed(String),
-    /// The remote party violated the protocol.
-    #[error("Protocol violation: {0}")]
-    ProtocolError(String),
-    /// GnuPG's home directory doesn't exist.
-    #[error("GnuPG's home directory ({0}) doesn't exist")]
-    GnuPGHomeMissing(PathBuf),
-    /// Unknown key.
-    #[error("Unknown key: {0}")]
-    UnknownKey(Keygrip),
-    /// No smartcards are connected.
-    #[error("No smartcards are connected")]
-    NoSmartcards,
-    /// The key already exists.
-    #[error("{0} already exists: {1}")]
-    KeyExists(Keygrip, String),
-}
-
 #[cfg(test)]
 mod test {
     use super::*;
@@ -1757,7 +1902,7 @@ mod test {
     #[test]
     fn signer_in_async_context() -> Result<()> {
         async fn async_context() -> Result<()> {
-            let ctx = match ipc::gnupg::Context::ephemeral() {
+            let ctx = match gnupg::Context::ephemeral() {
                 Ok(c) => c,
                 Err(e) => {
                     eprintln!("Failed to create ephemeral context: {}", e);
@@ -1793,7 +1938,7 @@ mod test {
     #[test]
     fn decryptor_in_async_context() -> Result<()> {
         async fn async_context() -> Result<()> {
-            let ctx = match ipc::gnupg::Context::ephemeral() {
+            let ctx = match gnupg::Context::ephemeral() {
                 Ok(c) => c,
                 Err(e) => {
                     eprintln!("Failed to create ephemeral context: {}", e);
@@ -1833,7 +1978,7 @@ mod test {
         match result {
             Ok(_agent) => panic!("Created an agent for a non-existent home!"),
             Err(err) => {
-                if let Some(Error::GnuPGHomeMissing(_)) = err.downcast_ref() {
+                if let Error::GnuPGHomeMissing(_) = err {
                     // Correct error.
                 } else {
                     panic!("Expected Error::GnuPGHomeMissing, got: {}", err);
@@ -1847,7 +1992,7 @@ mod test {
     // Test that we can import a key and sign a message.
     #[tokio::test]
     async fn import_key_and_sign() -> Result<()> {
-        let ctx = match ipc::gnupg::Context::ephemeral() {
+        let ctx = match gnupg::Context::ephemeral() {
             Ok(c) => c,
             Err(e) => {
                 eprintln!("Failed to create ephemeral context: {}", e);
@@ -1876,7 +2021,7 @@ mod test {
     async fn list_keys() -> Result<()> {
         use std::collections::HashSet;
 
-        let ctx = match ipc::gnupg::Context::ephemeral() {
+        let ctx = match gnupg::Context::ephemeral() {
             Ok(c) => c,
             Err(e) => {
                 eprintln!("Failed to create ephemeral context: {}", e);
@@ -1894,7 +2039,7 @@ mod test {
         let testy = import_testy_new(&mut agent).await?;
         let testy_keygrips = testy.keys().into_iter()
             .map(|ka| Keygrip::of(ka.key().mpis()))
-            .collect::<Result<HashSet<Keygrip>>>()
+            .collect::<Result<HashSet<Keygrip>, _>>()
             .expect("can compute keygrip");
         assert!(testy_keygrips.len() > 0);
 
@@ -1913,7 +2058,7 @@ mod test {
     async fn key_info() -> Result<()> {
         use std::collections::HashSet;
 
-        let ctx = match ipc::gnupg::Context::ephemeral() {
+        let ctx = match gnupg::Context::ephemeral() {
             Ok(c) => c,
             Err(e) => {
                 eprintln!("Failed to create ephemeral context: {}", e);
@@ -1928,7 +2073,7 @@ mod test {
         let testy = import_testy_new(&mut agent).await?;
         let testy_keygrips = testy.keys().into_iter()
             .map(|ka| Keygrip::of(ka.key().mpis()))
-            .collect::<Result<HashSet<Keygrip>>>()
+            .collect::<Result<HashSet<Keygrip>, _>>()
             .expect("can compute keygrip");
         assert!(testy_keygrips.len() > 0);
 
@@ -1963,7 +2108,7 @@ mod test {
         async fn test(file: &str, password: &str)
             -> Result<()>
         {
-            let ctx = match ipc::gnupg::Context::ephemeral() {
+            let ctx = match gnupg::Context::ephemeral() {
                 Ok(c) => c,
                 Err(e) => {
                     eprintln!("Failed to create ephemeral context: {}", e);
@@ -2001,10 +2146,11 @@ mod test {
                 Ok(()) => panic!("preset_passphrase should fail if \
                                   'allow-preset-passphrase' is not set."),
                 Err(err) => {
-                    if let Some(Error::OperationFailed(_)) = err.downcast_ref() {
+                    if let Error::Assuan(assuan::Error::OperationFailed(_)) = err {
                         // That's what we expect.
                     } else {
-                        panic!("Expected Error::OperationFailed, but got {}", err);
+                        panic!("Expected assuan::Error::OperationFailed, \
+                                but got {}", err);
                     }
                 }
             }
@@ -2104,4 +2250,43 @@ mod test {
 
         Ok(())
     }
+
+    // Test Agent::export
+    #[tokio::test]
+    async fn export() -> Result<()> {
+        use std::collections::HashSet;
+
+        let ctx = match gnupg::Context::ephemeral() {
+            Ok(c) => c,
+            Err(e) => {
+                eprintln!("Failed to create ephemeral context: {}", e);
+                eprintln!("Most likely, GnuPG isn't installed.");
+                eprintln!("Skipping test.");
+                return Ok(());
+            },
+        };
+
+        let mut agent = Agent::connect(&ctx).await?;
+
+        let testy = import_testy_new(&mut agent).await?;
+        let testy_keygrips = testy.keys().into_iter()
+            .map(|ka| Keygrip::of(ka.key().mpis()))
+            .collect::<Result<HashSet<Keygrip>, _>>()
+            .expect("can compute keygrip");
+        assert!(testy_keygrips.len() > 0);
+
+        for ka in testy.keys().secret() {
+            let keygrip = Keygrip::of(ka.key().mpis()).expect("has a keygrip");
+            match agent.export(ka.key().parts_as_public().clone()).await {
+                Ok(key) => {
+                    assert_eq!(ka.key(), &key);
+                }
+                Err(err) => {
+                    panic!("Exporting key {}: {}", keygrip, err);
+                }
+            }
+        }
+
+        Ok(())
+    }
 }
